use crate::error::Error;
use crate::utils;
use crate::utils::verify_download;
use log::info;
use url::Url;
use wicked_waifus_downloader_domain::game::{
    GameConfig, GameDefaultConfig, GameDefaultConfigV2Config, ResourcesData,
};
use wicked_waifus_downloader_domain::CdnInfo;
use wicked_waifus_downloader_lib::{
    download_cdn_asset_in_mem, request_chunk_resources, request_resources,
};

pub(crate) fn download_game(
    use_predownload: bool,
    use_chunked: bool,
    full: bool,
    game_config: GameConfig,
    proxy: Option<&str>,
    path: &str,
) -> Result<(), Error> {
    if use_predownload {
        game_config
            .predownload_switch
            .ok_or(Error::PreloadSwitchMissing)?;
        let _predownload = game_config.predownload.ok_or(Error::PreloadDataMissing)?;

        // TODO: Implement logic for download predownload
        unreachable!("Implement logic for download predownload");
    }
    match game_config.default {
        GameDefaultConfig::GameDefaultConfigV2(mut config) => {
            utils::sort_cdn_list(&mut config.cdn_list);
            let chunk_enabled = game_config.chunk_download_switch.unwrap_or(0) == 1;

            if full && use_chunked && chunk_enabled {
                download_full_game_chunked(&config.cdn_list, config.config, proxy, path)
            } else if full {
                download_full_game(&config.cdn_list, config.resources_data, proxy, path)
            } else {
                download_upgrade_game(&config.cdn_list, config.config, proxy, path)
            }
        }
        GameDefaultConfig::GameDefaultConfigV1(mut config) => {
            utils::sort_cdn_list(&mut config.cdn_list);
            download_full_game(&config.cdn_list, config.resources_data, proxy, path)
        }
    }
}

fn try_fetch_from_cdn<T>(
    cdn_list: &[CdnInfo],
    build_url: impl Fn(&str) -> Option<Url>,
    fetch_fn: impl Fn(&str) -> Result<T, Error>,
) -> Result<T, Error> {
    cdn_list
        .iter()
        .find_map(|cdn| {
            let url = build_url(&cdn.url)?;
            fetch_fn(url.as_str()).ok()
        })
        .ok_or(Error::CdnListExhausted)
}

fn download_full_game(
    cdn_list: &[CdnInfo],
    resources_data: ResourcesData,
    proxy: Option<&str>,
    path: &str,
) -> Result<(), Error> {
    let resources = try_fetch_from_cdn(
        cdn_list,
        |base| Url::parse(base).ok()?.join(&resources_data.resources).ok(),
        |url| Ok(request_resources(url, proxy)?),
    )?;

    let base_url = format!(
        "{}/",
        resources_data.resources_base_path.trim_end_matches('/')
    );
    let total_elements = resources.resource.len();
    for (i, resource) in resources.resource.iter().enumerate() {
        let expected_md5 = hex::decode(&resource.md5)?;

        let data = download_resource_file(
            cdn_list,
            &base_url,
            &resource.dest,
            resource.size,
            &expected_md5,
            i,
            total_elements,
            proxy,
        )?;

        utils::save_file_safe(path, &resource.dest, &data)?;
    }

    Ok(())
}

fn download_full_game_chunked(
    cdn_list: &[CdnInfo],
    resources_data: GameDefaultConfigV2Config,
    proxy: Option<&str>,
    path: &str,
) -> Result<(), Error> {
    let resources = try_fetch_from_cdn(
        cdn_list,
        |base| Url::parse(base).ok()?.join(&resources_data.index_file).ok(),
        |url| Ok(request_chunk_resources(url, proxy)?),
    )?;

    let base_url = format!("{}/", resources_data.base_url.trim_end_matches('/'));
    let total_elements = resources.resource.len();
    for (i, resource) in resources.resource.iter().enumerate() {
        let expected_md5 = hex::decode(&resource.md5)?;

        let data = if let Some(chunk_info) = &resource.chunk_infos {
            for _chunk in chunk_info {}
            unreachable!("Implement logic for chunk download");
        } else {
            download_resource_file(
                cdn_list,
                &base_url,
                &resource.dest,
                resource.size,
                &expected_md5,
                i,
                total_elements,
                proxy,
            )?
        };

        utils::save_file_safe(path, &resource.dest, &data)?;
    }

    Ok(())
}

fn download_upgrade_game(
    _cdn_list: &[CdnInfo],
    _resources_data: GameDefaultConfigV2Config,
    _proxy: Option<&str>,
    _path: &str,
) -> Result<(), Error> {
    unreachable!("implement upgrade/patch download");
}

fn download_resource_file(
    cdn_list: &[CdnInfo],
    base_url: &str,
    dest: &str,
    expected_length: u64,
    expected_md5: &[u8],
    cnt: usize,
    total_elements: usize,
    proxy: Option<&str>,
) -> Result<Vec<u8>, Error> {
    try_fetch_from_cdn(
        cdn_list,
        |base| Url::parse(base).ok()?.join(base_url).ok()?.join(dest).ok(),
        |url| {
            info!("[{}/{}] Download started: {}", cnt + 1, total_elements, url);
            let data = download_cdn_asset_in_mem(url, proxy)?;
            verify_download(&data, expected_length, expected_md5)?;
            Ok(data)
        },
    )
}
