use crate::error::Error;
use crate::utils;
use crate::utils::verify_download;
use log::{info, warn};
use url::Url;
use wicked_waifus_downloader_domain::launcher::LauncherConfig;
use wicked_waifus_downloader_lib::{download_cdn_asset_in_mem, InstallerInfo};

pub(crate) fn download_launcher(
    launcher_config: LauncherConfig,
    proxy: Option<&str>,
    path: &str,
) -> Result<(), Error> {
    let mut info = InstallerInfo::from(launcher_config);
    let expected_md5 = hex::decode(&info.expected_md5)?;
    utils::sort_cdn_list(&mut info.cdn_list);

    info.cdn_list.iter().find_map(|cdn| {
        let Ok(url) = Url::parse(&cdn.url).and_then(|base| base.join(&info.installer)) else {
            warn!("Invalid URL in CDN list: {}", cdn.url);
            return None;
        };

        info!("Download started: {}", url);

        let data = match download_cdn_asset_in_mem(url.as_str(), proxy) {
            Ok(data) => data,
            Err(err) => {
                warn!("Download failed from {}: {:?}", url, err);
                return None;
            }
        };

        if let Err(e) = verify_download(&data, info.expected_length, &expected_md5) {
            warn!("Verification failed for {}: {:?}", url, e);
            return None;
        }

        if let Err(e) = utils::save_file_safe(path, "installer.zip", &data) {
            warn!("Failed to save file: {:?}", e);
            return None;
        }

        Some(())
    }).ok_or(Error::CdnListExhausted)
}
