use indicatif::{MultiProgress, ProgressBar};
use log::{info, warn};
use std::fmt::Write;
use std::sync::OnceLock;
use serde::de::DeserializeOwned;
use thiserror::Error;
use ureq::http::Response;
use ureq::tls::{RootCerts, TlsConfig};
use ureq::{Agent, Body, Proxy};
use wicked_waifus_downloader_domain::game::GameConfig;
use wicked_waifus_downloader_domain::launcher::LauncherConfig;
use wicked_waifus_downloader_domain::resources::{ChunkedResources, Resources};
use wicked_waifus_downloader_domain::CdnInfo;

#[derive(Error, Debug)]
pub enum Error {
    #[error("UREQ Error {0}")]
    Ureq(#[from] ureq::Error),
    #[error("ToStr Error {0}")]
    ToStr(#[from] ureq::http::header::ToStrError),
    #[error("ParseInt Error {0}")]
    ParseInt(#[from] std::num::ParseIntError),
    #[error("Io Error {0}")]
    Io(#[from] std::io::Error),
    #[error("Content-Length header missing")]
    ContentLengthHeaderMissing,
}

pub struct InstallerInfo {
    pub cdn_list: Vec<CdnInfo>,
    pub installer: String,
    pub expected_md5: String,
    pub expected_length: u64,
}

impl From<LauncherConfig> for InstallerInfo {
    fn from(config: LauncherConfig) -> Self {
        match config {
            LauncherConfig::LauncherConfigV1(cfg) => {
                let default = cfg.default;
                let common = default.common;
                Self {
                    cdn_list: common.cdn_list,
                    installer: default.installer,
                    expected_md5: default.installer_md5,
                    expected_length: default.installer_size,
                }
            }
            LauncherConfig::LauncherConfigV2(cfg) => {
                let default = cfg.default;
                let common = default.common;
                let resource = default.resource;
                Self {
                    cdn_list: common.cdn_list,
                    installer: resource.path,
                    expected_md5: resource.md5,
                    expected_length: resource.size,
                }
            }
        }
    }
}

static MULTI_PROGRESS_BAR: OnceLock<MultiProgress> = OnceLock::new();

pub fn enable_progress_bar(multi: MultiProgress) {
    if MULTI_PROGRESS_BAR.set(multi).is_err() {
        warn!("MultiProgress already set, ignoring.");
    }
}

macro_rules! define_request_fn {
    ($name:ident, $ty:ty) => {
        pub fn $name(url: &str, proxy: Option<&str>) -> Result<$ty, Error> {
            request_json::<$ty>(url, proxy)
        }
    };
}

define_request_fn!(request_launcher, LauncherConfig);
define_request_fn!(request_game, GameConfig);
define_request_fn!(request_resources, Resources);
define_request_fn!(request_chunk_resources, ChunkedResources);

pub fn download_cdn_asset_in_mem(url: &str, proxy: Option<&str>) -> Result<Vec<u8>, Error> {
    let (mut res, len) = prepare_download(url, proxy)?;
    let mut bytes = Vec::with_capacity(len);
    download_internal(url, &mut res, len as u64, &mut bytes)?;
    Ok(bytes)
}

pub fn download_cdn_asset_writer<W: std::io::Write>(
    url: &str,
    proxy: Option<&str>,
    writer: &mut W,
) -> Result<(), Error> {
    let (mut res, len) = prepare_download(url, proxy)?;
    download_internal(url, &mut res, len as u64, writer)
}

fn get_agent(proxy: Option<&str>) -> Result<Agent, Error> {
    Ok(Agent::config_builder()
        .tls_config(
            TlsConfig::builder()
                .root_certs(RootCerts::PlatformVerifier)
                .build(),
        )
        .proxy(proxy.map(Proxy::new).transpose()?)
        .build()
        .new_agent())
}

#[inline(always)]
fn request_json<T: DeserializeOwned>(url: &str, proxy: Option<&str>) -> Result<T, Error> {
    Ok(get_agent(proxy)?
        .get(url)
        .call()?
        .body_mut()
        .read_json::<T>()?)
}

fn prepare_download(url: &str, proxy: Option<&str>) -> Result<(Response<Body>, usize), Error> {
    let res = get_agent(proxy)?.get(url).call()?;
    let content_length = res
        .headers()
        .get("Content-Length")
        .ok_or(Error::ContentLengthHeaderMissing)?
        .to_str()?
        .parse::<usize>()?;
    Ok((res, content_length))
}

fn create_progress_bar(len: u64, multi: &MultiProgress) -> ProgressBar {
    let pb = multi.add(ProgressBar::new(len));
    pb.set_style(
        indicatif::ProgressStyle::with_template(
            "{spinner:.green} [{elapsed_precise}] [{wide_bar:.cyan/blue}] {bytes}/{total_bytes} ({eta})"
        )
            .unwrap()
            .with_key("eta", |state: &indicatif::ProgressState, w: &mut dyn Write| {
                write!(w, "{:.1}s", state.eta().as_secs_f64()).unwrap();
            })
            .progress_chars("#>-")
    );
    pb
}

fn download_internal<W: std::io::Write>(
    url: &str,
    res: &mut Response<Body>,
    len: u64,
    writer: &mut W,
) -> Result<(), Error> {
    if let Some(multi) = MULTI_PROGRESS_BAR.get() {
        let pb = create_progress_bar(len, multi);
        std::io::copy(&mut pb.wrap_read(res.body_mut().as_reader()), writer)?;
        pb.finish();
        multi.remove(&pb);
    } else {
        std::io::copy(&mut res.body_mut().as_reader(), writer)?;
    }

    info!("Download completed: {}", url);
    Ok(())
}
